-------------------------------------------------------------------------------
-- FILE NAME   : BAUDGEN.VHD
-- PROJECT     : UART
-- VERSION     : A
-- AUTHOR      : M. S. HARVEY
-- SIM/SYN     : Synthesis
-- DESCRIPTION : Baud Rate Generator
-- DEPENDENCIES: IEEE.STD_LOGIC_1164; IEEE.NUMERIC_STD
-- LIMITATIONS :
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- CHANGE HISTORY.
-- REV.           DATE        DESCRIPTION					
--  A             02/12/99    Initial design.
-------------------------------------------------------------------------------
                                   
------------------------------------------------------------------------------
-- NOTE! : The generic 'accum_size' must be set to a value which is equal to or
-- greater than the width of the 'n' input which is 16 bits.
-- The actual accumulator generated will be 1 bit wider than the value of accum_size.
------------------------------------------------------------------------------

                                   
-- IEEE library
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;


-- accum_size must be equal to or greater than width of n
ENTITY baudgen IS
  GENERIC (
    accum_size : integer:=16);                       -- accumulator size (in bits)
  PORT(
    hclk     : IN  std_logic;                        -- host clock
    reseth   : IN  std_logic;                        -- async reset
    n        : IN  std_logic_vector(15 DOWNTO 0);    -- BRSEL1,0
    bclkx16  : OUT std_logic                         -- baud clock x 16
    );
END ENTITY baudgen;



ARCHITECTURE arc_baudgen OF baudgen IS

  SIGNAL s_q : unsigned(accum_size DOWNTO 0);        -- accumulator register
  SIGNAL s_e : unsigned(accum_size DOWNTO 0);        -- zero extended n


BEGIN

  -- s_e signal is the 'n' input with zero padding to be same width as accumulator
  s_e(n'LENGTH-1 DOWNTO 0)        <= unsigned(n);
  s_e(accum_size DOWNTO n'LENGTH) <= (OTHERS => '0');

  -- accumulator
  baud_accum : PROCESS (reseth,hclk)
  BEGIN
    IF (reseth='1') THEN
      s_q <= (OTHERS => '0');
    ELSIF (hclk'EVENT AND hclk='1') THEN
      s_q <= s_q + s_e;
    END IF;
  END PROCESS;

  bclkx16 <= std_logic(s_q(accum_size));              -- the x16 baud clock
  

END ARCHITECTURE arc_baudgen;
------------------------------------------------------------------------------
-- END OF FILE : BAUDGEN.VHD
------------------------------------------------------------------------------