-------------------------------------------------------------------------------
-- FILE NAME   : HOST.VHD
-- PROJECT     : UART
-- VERSION     : A
-- AUTHOR      : M. S. HARVEY
-- SIM/SYN     : Synthesis
-- DESCRIPTION : Host Interface & registers
-- DEPENDENCIES: IEEE.STD_LOGIC_1164
-- LIMITATIONS :
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- CHANGE HISTORY.
-- REV.           DATE        DESCRIPTION					
--  A             05/12/99    Initial design.
-------------------------------------------------------------------------------


-- IEEE library
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;


ENTITY host IS
  PORT(
    hclk       : IN  std_logic;                      -- host clock
    reseth     : IN  std_logic;                      -- async reset
    cs         : IN  std_logic;                      -- chip select
    rw         : IN  std_logic;                      -- read/write
    addr       : IN  std_logic_vector(2 DOWNTO 0);   -- host address
    data_in    : IN  std_logic_vector(7 DOWNTO 0);   -- host data input for writes
    rxdata     : IN  std_logic_vector(7 DOWNTO 0);   -- received data (from rx hold reg)
    frm_err    : IN  std_logic;                      -- frame error from receiver
    pty_err    : IN  std_logic;                      -- parity error from receiver
    ovr_err    : IN  std_logic;                      -- overrun error from receiver
    txempty    : IN  std_logic;                      -- transmitter hold & shift reg empty
    txrdy      : IN  std_logic;                      -- transmitter hold reg empty
    rxrdy      : IN  std_logic;                      -- receiver data ready in hold reg
    rxfull     : IN  std_logic;                      -- receiver hold & shift reg full
    n          : OUT std_logic_vector(15 DOWNTO 0);  -- brsel0,1 concatenated
    ptymode    : OUT std_logic_vector(1 DOWNTO 0);   -- parity mode
    chanmode   : OUT std_logic_vector(1 DOWNTO 0);   -- channel operating mode
    txen       : OUT std_logic;                      -- transmitter enable
    txctsen    : OUT std_logic;                      -- transmitter inhibit by cts enable
    rxrtsen    : OUT std_logic;                      -- receiver rts enable
    rxen       : OUT std_logic;                      -- receiver enable
    txholdwe   : OUT std_logic;                      -- transmitter hold reg write enable
    rxholdrd   : OUT std_logic;                      -- receiver hold reg read enable
    rxstatusrst: OUT std_logic;                      -- status register reset (rx flags)
    data_out   : OUT std_logic_vector(7 DOWNTO 0);   -- host data output for reads
    dataoe     : OUT std_logic;                      -- host data output enable for reads
    irq        : OUT std_logic;                      -- interrupt request to host
    ack        : OUT std_logic                       -- transfer acknowledge to host
        );
END ENTITY host;



ARCHITECTURE arc_host OF host IS

-- state machine encoding (gray)
CONSTANT idle : std_logic_vector(1 DOWNTO 0):= "00"; 
CONSTANT s1   : std_logic_vector(1 DOWNTO 0):= "01"; 
CONSTANT s2   : std_logic_vector(1 DOWNTO 0):= "11"; 
CONSTANT s3   : std_logic_vector(1 DOWNTO 0):= "10";

-- register addresses
CONSTANT c_ctrladdr   : std_logic_vector(2 DOWNTO 0):= o"0";     -- control reg
CONSTANT c_brsel0addr : std_logic_vector(2 DOWNTO 0):= o"1";     -- BRSEL0 reg
CONSTANT c_brsel1addr : std_logic_vector(2 DOWNTO 0):= o"2";     -- BRSEL1 reg
CONSTANT c_statusaddr : std_logic_vector(2 DOWNTO 0):= o"3";     -- status reg
CONSTANT c_intmskaddr : std_logic_vector(2 DOWNTO 0):= o"4";     -- interrupt mask reg
CONSTANT c_holdaddr   : std_logic_vector(2 DOWNTO 0):= o"5";     -- tx hold (write)/rx hold (read)

-- internal signals
SIGNAL s_prestate    : std_logic_vector(1 DOWNTO 0);            -- present state vector
SIGNAL s_nextstate   : std_logic_vector(1 DOWNTO 0);            -- next state vector
SIGNAL s_nextack     : std_logic;                               -- next ack
SIGNAL s_nextmastwe  : std_logic;                               -- next master write enable
SIGNAL s_nextdataoe  : std_logic;                               -- next data out enable
SIGNAL s_nextrxholdrd: std_logic;                               -- next rx hold reg read
SIGNAL s_rxholdrd    : std_logic;                               -- rx hold reg read

-- registers
SIGNAL s_ctrlreg   : std_logic_vector(7 DOWNTO 0);              -- control
SIGNAL s_brsel0reg : std_logic_vector(7 DOWNTO 0);              -- BRSEL0
SIGNAL s_brsel1reg : std_logic_vector(7 DOWNTO 0);              -- BRSEL1
SIGNAL s_statusreg : std_logic_vector(7 DOWNTO 0);              -- status
SIGNAL s_intmskreg : std_logic_vector(1 DOWNTO 0);              -- interrupt mask

-- register write enables
SIGNAL s_masterwe  : std_logic;                                 -- master write enable
SIGNAL s_ctrlwe    : std_logic;                                 -- control reg write enable
SIGNAL s_brsel0we  : std_logic;                                 -- BRSEL0 reg write enable
SIGNAL s_brsel1we  : std_logic;                                 -- BRSEL1 reg write enable
SIGNAL s_intmskwe  : std_logic;                                 -- interrupt mask reg write enable

-- parity modes
CONSTANT c_noparity   : std_logic_vector(1 DOWNTO 0):= "00"; 
CONSTANT c_evenparity : std_logic_vector(1 DOWNTO 0):= "01"; 
CONSTANT c_oddparity  : std_logic_vector(1 DOWNTO 0):= "10";

-- operating modes 
CONSTANT c_normal    : std_logic_vector(1 DOWNTO 0):= "00"; 
CONSTANT c_intloop   : std_logic_vector(1 DOWNTO 0):= "01"; 
CONSTANT c_autoecho  : std_logic_vector(1 DOWNTO 0):= "10"; 


BEGIN

  ----------------------------------------------------------------------------
  -- STATE MACHINE
  ----------------------------------------------------------------------------
  
  sync_host : PROCESS (reseth,hclk)
    BEGIN
      IF (reseth='1') THEN
        s_prestate <= idle;
        ack        <= '1';
        s_masterwe <= '0';
        dataoe     <= '1';
        s_rxholdrd <= '0';
      ELSIF (rising_edge(hclk)) THEN
        s_prestate <= s_nextstate;
        ack        <= s_nextack;
        s_masterwe <= s_nextmastwe;
        dataoe     <= s_nextdataoe;
        s_rxholdrd <= s_nextrxholdrd;
      END IF;
  END PROCESS;
  
  
  -- next state logic
  comb_host : PROCESS (s_prestate,cs)
    BEGIN
      CASE s_prestate IS
        WHEN idle =>
          IF (cs='0') THEN                    -- start when cs active
            s_nextstate <= s1;
          ELSE
            s_nextstate <= idle;
          END IF;
        WHEN s1 =>
          s_nextstate <= s2;
        WHEN s2 =>
          s_nextstate <= s3;
        WHEN s3 =>
          s_nextstate <= idle;
        WHEN OTHERS =>
          s_nextstate <= (OTHERS => '-');     -- don't care for any undefined states
      END CASE;
  END PROCESS;

  
  -- output logic
  comb_outs : PROCESS (s_nextstate,rw,addr)
    BEGIN
      CASE s_nextstate IS
        WHEN idle =>
          s_nextack      <= '1';
          s_nextmastwe   <= '0';
          s_nextdataoe   <= '1';
          s_nextrxholdrd <= '0';            
        WHEN s1 =>
          s_nextack      <= '1';
          s_nextmastwe   <= '0';
          s_nextdataoe   <= '1';
          s_nextrxholdrd <= '0';            
        WHEN s2 =>
          s_nextack    <= '1';
          s_nextmastwe <= NOT(rw);    -- master write strobe asserted for writes
          s_nextdataoe <= NOT(rw);    -- data output enabled for reads
          IF (rw='1' AND addr=c_holdaddr) THEN
            s_nextrxholdrd <= '1';
          ELSE
            s_nextrxholdrd <= '0';            
          END IF;
        WHEN s3 =>
          s_nextack    <= '0';        -- ack asserted
          s_nextmastwe <= '0';
          s_nextdataoe <= NOT(rw);    -- data output enabled for reads
          IF (rw='1' AND addr=c_holdaddr) THEN
            s_nextrxholdrd <= '1';
          ELSE
            s_nextrxholdrd <= '0';            
          END IF;
        WHEN OTHERS =>
          s_nextack      <= '1';
          s_nextmastwe   <= '0';
          s_nextdataoe   <= '1';
          s_nextrxholdrd <= '0';            

      END CASE;
  END PROCESS;

    
  ------------------------------------------------------------------------------
  -- REGISTER WRITES
  ------------------------------------------------------------------------------

  -- control register write enable
  s_ctrlwe <= '1' WHEN (s_masterwe='1' AND addr=c_ctrladdr) ELSE '0';
  
  -- baud rate select #0 register write enable
  s_brsel0we <= '1' WHEN (s_masterwe='1' AND addr=c_brsel0addr) ELSE '0';
  
  -- baud rate select #1 register write enable
  s_brsel1we <= '1' WHEN (s_masterwe='1' AND addr=c_brsel1addr) ELSE '0';
  
  -- status register write enable (write clears rx flags)
  rxstatusrst<= '1' WHEN (s_masterwe='1' AND addr=c_statusaddr) ELSE '0';
  
  -- interrupt mask register write enable
  s_intmskwe <= '1' WHEN (s_masterwe='1' AND addr=c_intmskaddr) ELSE '0';

  -- tx holding register  write enable (write only)
  -- can't write tx holding reg in autoecho mode or if transmitter disabled
  txholdwe <= '1' WHEN (s_masterwe='1' AND addr=c_holdaddr AND
                        s_ctrlreg(7 DOWNTO 6)/=c_autoecho AND s_ctrlreg(0)='1')
                  ELSE '0';


  -- writes
  write_regs : PROCESS (reseth,hclk)
    BEGIN
      IF (reseth='1') THEN
        s_ctrlreg   <= (OTHERS => '0');
        s_brsel1reg <= (OTHERS => '0');
        s_brsel0reg <= (OTHERS => '0');
        s_intmskreg <= (OTHERS => '0');
        
      ELSIF (rising_edge(hclk)) THEN
        
        IF (s_ctrlwe='1') THEN         -- control write
          s_ctrlreg <= data_in;
        END IF;
        
        IF (s_brsel0we='1') THEN       -- BRSEL0 write
          s_brsel0reg <= data_in;
        END IF;
        
        IF (s_brsel1we='1') THEN       -- BRSEL0 write
          s_brsel1reg <= data_in;
        END IF;
                
        IF (s_intmskwe='1') THEN       -- interrupt mask write
          s_intmskreg <= data_in(1 DOWNTO 0);
        END IF;
                
      END IF;
  END PROCESS;

   
  ------------------------------------------------------------------------------
  -- REGISTER READS
  ------------------------------------------------------------------------------

  -- multiplex registers onto data output bus
  WITH addr SELECT
    data_out <= s_ctrlreg              WHEN c_ctrladdr,       -- control reg
                s_brsel0reg            WHEN c_brsel0addr,     -- brate select #0 reg
                s_brsel1reg            WHEN c_brsel1addr,     -- brate select #1 reg
                s_statusreg            WHEN c_statusaddr,     -- status reg
                "000000" & s_intmskreg WHEN c_intmskaddr,     -- interrupt mask
                rxdata                 WHEN OTHERS;           -- received data (rx hold reg)
  
 
  -- rx hold register read (see state machine)
  rxholdrd <= s_rxholdrd;
  
  -- concatenate flags to make status register
  s_statusreg <= '0' & frm_err & pty_err & ovr_err & txempty & txrdy & rxfull & rxrdy;
  
  
  ----------------------------------------------------------------------------
  -- INTERRUPT
  ----------------------------------------------------------------------------
  -- the interrupt request will be generated for the following conditions :
  --   1. data ready in receiver hold reg if interrupt not masked.
  --   2. transmitter hold reg empty if transmitter enabled and interrupt not masked.
  -- Note that transmitter hold reg empty interrupt will occur when the transmitter 
  -- is first enabled. It can only be cleared by writing new data to the buffer or
  -- by disabling the transmitter by setting txen to 0. It will not be generated
  -- in autoecho mode.
   
  irq<='0' WHEN ( (rxrdy='1' AND s_intmskreg(0)='1') OR
                  (txrdy='1' AND s_intmskreg(1)='1' AND s_ctrlreg(0)='1') )
           ELSE '1';
  
  
  ------------------------------------------------------------------------------
  -- CONTROL REGISTER OUTPUTS
  ------------------------------------------------------------------------------
  n        <= s_brsel1reg & s_brsel0reg;   -- baud rate select regs concatenated
  txen     <= s_ctrlreg(0);                -- transmitter enable
  rxen     <= s_ctrlreg(1);                -- receiver enable
  ptymode  <= s_ctrlreg(3 DOWNTO 2);       -- parity mode           
  rxrtsen  <= s_ctrlreg(4);                -- receiver cts enable
  txctsen  <= s_ctrlreg(5);                -- transmitter inhibit by cts
  chanmode <= s_ctrlreg(7 DOWNTO 6);       -- channel operating mode           
  

  

END ARCHITECTURE arc_host;
-------------------------------------------------------------------------------
-- END OF FILE : HOST.VHD
-------------------------------------------------------------------------------
