-------------------------------------------------------------------------------
-- FILE NAME   : RECEIVE.VHD
-- PROJECT     : UART
-- VERSION     : A
-- AUTHOR      : M. S. HARVEY
-- SIM/SYN     : Synthesis
-- DESCRIPTION : Receiever
-- DEPENDENCIES: IEEE.STD_LOGIC_1164; IEEE.NUMERIC_STD
-- LIMITATIONS :
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- CHANGE HISTORY.
-- REV.           DATE        DESCRIPTION					
--  A             06/12/99    Initial design.
-------------------------------------------------------------------------------


-- IEEE library
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;


ENTITY receive IS
  PORT(
    hclk       : IN  std_logic;                      -- host clock
    reseth     : IN  std_logic;                      -- async reset
    bclkx16    : IN  std_logic;                      -- x16 baud rate clock
    rxen       : IN  std_logic;                      -- receiver enable
    rxd        : IN  std_logic;                      -- received data
    rxrtsen    : IN  std_logic;                      -- rts enable
    rxholdrd   : IN  std_logic;                      -- rx hold reg read enable
    rxstatusrst: IN  std_logic;                      -- status register clear command
    ptymode    : IN  std_logic_vector(1 DOWNTO 0);   -- parity mode
    rxdata     : OUT std_logic_vector(7 DOWNTO 0);   -- data output for reads from rx hold reg
    rts        : OUT std_logic;                      -- ready to send
    rxrdy      : OUT std_logic;                      -- receiver hold reg full
    rxfull     : OUT std_logic;                      -- receiver hold & shift reg full
    frm_err    : OUT std_logic;                      -- frame error to host
    pty_err    : OUT std_logic;                      -- parity error to host
    ovr_err    : OUT std_logic                       -- overrun error to host
      );
END ENTITY receive;



ARCHITECTURE arc_receive OF receive IS


  -- internal signals
  SIGNAL s_pstate     : std_logic_vector(3 DOWNTO 0); -- state machine present state vector
  SIGNAL s_nstate     : std_logic_vector(3 DOWNTO 0); -- state machine next state vector
  SIGNAL s_rxaccept   : std_logic;                    -- receiver has accepted frame (i.e shifting in)
  SIGNAL s_rxddelay   : std_logic;                    -- delayed version of received data
  SIGNAL s_rxclken    : std_logic;                    -- receiver bit rate clock enable
  SIGNAL s_rxshiften  : std_logic;                    -- receiver shift register shift enable
  SIGNAL s_rxshiftreg : std_logic_vector(7 DOWNTO 0); -- receiver shift register
  SIGNAL s_rxholdreg  : std_logic_vector(7 DOWNTO 0); -- receiver holding register
  SIGNAL s_rxpty      : std_logic;                    -- calculated receiver parity
  SIGNAL s_ovr_err    : std_logic;                    -- internal version of overrun error
  SIGNAL s_frmerr     : std_logic;                    -- internal version of frame error
  SIGNAL s_rxrdy      : std_logic;                    -- receiver ready flag
  SIGNAL s_rstrxrdy   : std_logic;                    -- clear rxrdy on hold reg read 
  SIGNAL s_rxholdrd1  : std_logic;                    -- rx hold reg rd delayed by 1 hclk period
  SIGNAL s_bclkcount  : unsigned(3 DOWNTO 0);
  SIGNAL s_rxshiftfull: std_logic;                    -- rx shift register full
  SIGNAL s_rxstart    : std_logic;                    -- testing start bit
  
  -- state machine encoding (explicit, binary)
  CONSTANT start : std_logic_vector(3 DOWNTO 0) := x"0";  
  CONSTANT d0    : std_logic_vector(3 DOWNTO 0) := x"4";  
  CONSTANT d1    : std_logic_vector(3 DOWNTO 0) := x"c";  
  CONSTANT d2    : std_logic_vector(3 DOWNTO 0) := x"8";  
  CONSTANT d3    : std_logic_vector(3 DOWNTO 0) := x"9";  
  CONSTANT d4    : std_logic_vector(3 DOWNTO 0) := x"b";  
  CONSTANT d5    : std_logic_vector(3 DOWNTO 0) := x"f";  
  CONSTANT d6    : std_logic_vector(3 DOWNTO 0) := x"7";  
  CONSTANT d7    : std_logic_vector(3 DOWNTO 0) := x"5";  
  CONSTANT pty   : std_logic_vector(3 DOWNTO 0) := x"3";  
  CONSTANT stop  : std_logic_vector(3 DOWNTO 0) := x"1";  
    
  -- parity modes
  CONSTANT c_noparity   : std_logic_vector(1 DOWNTO 0):= "00"; 
  CONSTANT c_evenparity : std_logic_vector(1 DOWNTO 0):= "01"; 
  CONSTANT c_oddparity  : std_logic_vector(1 DOWNTO 0):= "10";
  
  -- operating modes 
  CONSTANT c_normal    : std_logic_vector(1 DOWNTO 0):= "00"; 
  CONSTANT c_intloop   : std_logic_vector(1 DOWNTO 0):= "01"; 
  CONSTANT c_autoecho  : std_logic_vector(1 DOWNTO 0):= "10"; 
  

BEGIN


  -- rts output is negated (high) to block external transmitters when the
  -- receiver is not enabled. It will only go active (low) when rts control mode
  -- is not enabled or when rx hold register is empty if rts control mode is enabled.
  rts <= '1' WHEN (rxen='0') ELSE
         '0' WHEN (rxrtsen='0') ELSE
          s_rxrdy;
  
  ------------------------------------------------------------------------------
  -- RECEIVER BITCLOCK
  ------------------------------------------------------------------------------
  -- A 4 bit counter which will be reset to "0000" after a global asynchronous reset.
  -- It will reset to "0000" synchronously if it detects a low on the serial 
  -- data input while waiting for the start bit. This ensures that the bclkx16
  -- rising edge is positioned correctly at the start of each received frame.
  
  rx_bclkgen : PROCESS (bclkx16,reseth)
      
  BEGIN
    IF (reseth='1') THEN
      s_rxstart  <= '0';
      s_rxddelay <= '1';
      s_bclkcount<=(OTHERS=>'0');
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      
      -- delayed version of rxd, used to detect falling edge on rxd
      s_rxddelay <= rxd;
      
      -- reset if received data goes low while waiting for start bit
      IF (s_pstate=start AND rxd='0' AND s_rxddelay='1') THEN
        s_bclkcount<=(OTHERS=>'0');
      ELSE
        s_bclkcount<=s_bclkcount + 1;      -- increment       
      END IF;
      
      -- start bit status. goes hi on rxd falling edge & stays
      -- hi if rxd remains low for at least half a bit period
      IF ((s_rxstart='0' AND s_pstate=start AND rxd='0' AND s_rxddelay='1') OR
          (s_rxstart='1' AND s_pstate=start AND rxd='0'))
      THEN
        s_rxstart <= '1';
      ELSE
        s_rxstart <= '0';
      END IF; 
      
    END IF;
        
  END PROCESS;
  
  s_rxclken <= '1' WHEN (s_bclkcount=x"7") ELSE '0';

  ------------------------------------------------------------------------------
  -- RECEIVER STATE MACHINE
  ------------------------------------------------------------------------------
  -- Two process state machine.
  -- the receiver state machine will be activated for every frame received.
  
  -- syncronous process
  rx_sync : PROCESS (reseth,bclkx16)
    BEGIN
      IF (reseth='1') THEN
        s_pstate <= start;
      ELSIF (bclkx16'event AND bclkx16='1') THEN
        IF (s_rxclken='1') THEN
          s_pstate <= s_nstate;
        END IF;
      END IF;      
  END PROCESS rx_sync; 

  -- combinatorial process (next state logic)
  rx_comb : PROCESS (s_pstate,s_rxstart,ptymode)
    BEGIN
      CASE s_pstate IS
        WHEN start =>
          IF (s_rxstart='1') THEN      -- start bit found
            s_nstate <= d0;
          ELSE
            s_nstate <= start;
          END IF; 
        WHEN d0 =>
          s_nstate <= d1;
        WHEN d1 =>
          s_nstate <= d2;
        WHEN d2 =>
          s_nstate <= d3;
        WHEN d3 =>
          s_nstate <= d4;
        WHEN d4 =>
          s_nstate <= d5;
        WHEN d5 =>
          s_nstate <= d6;
        WHEN d6 =>
          s_nstate <= d7;
        WHEN d7 =>
          IF (ptymode=c_noparity) THEN
            s_nstate <= stop;
          ELSE
            s_nstate <= pty;
          END IF; 
        WHEN pty =>
          s_nstate <= stop;
        WHEN stop =>
          s_nstate <= start;
        WHEN OTHERS =>
          s_nstate <= (OTHERS => '-');
      END CASE; 
  END PROCESS rx_comb; 

  
  -- If the start bit is accepted as ok,(ie. stays low for 8 rxclk periods)
  -- then bclkx16 will set s_rxaccept high in the middle of the start bit.
  -- It will be set low again in the middle of the stop bit.
  rx_waitstart : PROCESS (bclkx16,reseth,rxen)
  BEGIN
    IF (reseth='1' OR rxen='0') THEN
      s_rxaccept  <='0';
      s_rxshiften <= '0';
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_rxclken='1') THEN

        -- s_rxaccept goes hi if start bit ok & shift reg not full
        -- hi until stop bit
        IF ((s_pstate=start AND s_rxstart='1' AND s_rxshiftfull='0') OR
            (s_rxaccept='1' AND s_pstate/=stop))
        THEN
          s_rxaccept<='1';
        ELSE
          s_rxaccept<='0';
        END IF;
        
        -- s_rxshiften goes hi if start bit ok & shift reg not full
        -- only hi during data bits
        IF ((s_pstate=start AND s_rxstart='1' AND s_rxshiftfull='0') OR
            (s_rxshiften='1' AND s_pstate/=d7))
        THEN
          s_rxshiften <='1';
        ELSE
          s_rxshiften <='0';
        END IF;
      
      END IF;
      
    END IF;
    
  END PROCESS;
  
  ------------------------------------------------------------------------------
  -- RECEIVER SHIFT REGISTER
  ------------------------------------------------------------------------------
  -- Shifts in the serial data. It is cleared to all 1's after a global reset.
  -- Note that the shift register will not accept new data if it is already full- an
  -- overrun error will be generated but the data in the shift register will not be
  -- overwritten.
  rx_shiftreg : PROCESS (bclkx16,reseth,rxen)
  BEGIN
    IF (reseth='1' OR rxen='0') THEN
      s_rxshiftreg  <= (OTHERS=>'1');
      s_rxshiftfull <= '0';
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_rxclken='1') THEN

        IF (s_rxshiften='1') THEN
          s_rxshiftreg <= s_rxddelay & s_rxshiftreg(7 DOWNTO 1); -- shift data in
        ELSE
          s_rxshiftreg <=s_rxshiftreg;                    -- hold current value.
        END IF;
        
        -- shift register is full if shift finishes and rx hold register is full
        IF ((s_pstate=stop AND s_rxaccept='1' AND s_rxrdy='1') OR
            (s_rxshiftfull='1' AND s_rxrdy='1'))      
        THEN
          s_rxshiftfull<='1';
        ELSE
          s_rxshiftfull<='0';
        END IF; 
        
      END IF;
      
    END IF;
    
  END PROCESS;
  
  rxfull <= s_rxshiftfull;
  ------------------------------------------------------------------------------
  -- RECEIVER READY FLAG
  ------------------------------------------------------------------------------
  -- s_rxrdy set to 1 when serial data is transferred to receiver holding register
  -- (s_rxholdreg) and cleared on read of receiver holding register or global reset
  rx_rdyflag : PROCESS (bclkx16,reseth,s_rstrxrdy,rxen)
  BEGIN
    IF (reseth='1' OR rxen='0' OR s_rstrxrdy='1') THEN
      s_rxrdy<='0';
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_rxclken='1') THEN
    
        IF ((s_pstate=stop AND s_rxaccept='1' AND s_rxrdy='0') OR
            (s_rxshiftfull='1' AND s_rxrdy='0'))
        THEN
          s_rxrdy<= '1';
        END IF;
        
      END IF;
      
    END IF;
    
  END PROCESS;
  
  rxrdy <= s_rxrdy;

  
  ------------------------------------------------------------------------------
  -- RECEIVER HOLDING REGISTER
  ------------------------------------------------------------------------------
  -- transfer of data to rx holding register when receive is finished if rx hold
  -- is empty (s_rxrdy='0') and data is pending in the shift register.
  -- the data in the rx holding register will not be overwritten if an overrun occurs.
  rx_hold : PROCESS (bclkx16,reseth,rxen)
  BEGIN
    IF (reseth='1' OR rxen='0') THEN
      s_rxholdreg <= (OTHERS=>'1');
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_rxclken='1') THEN

        IF ((s_pstate=stop AND s_rxaccept='1' AND s_rxrdy='0') OR
            (s_rxshiftfull='1' AND s_rxrdy='0'))
        THEN
          s_rxholdreg <= s_rxshiftreg(7 DOWNTO 0);
        END IF;
        
      END IF;
        
    END IF;
    
  END PROCESS;

  -- detect read of receiver holding register & generate reset for receiver
  -- ready flag (s_rxrdy)
  rx_clearrxrdy: PROCESS (hclk,reseth,rxen)
  BEGIN
      IF (reseth='1' OR rxen='0') THEN 
         s_rxholdrd1 <= '0';
      ELSIF (hclk'event AND hclk='1') THEN 
      
         s_rxholdrd1 <= rxholdrd;
                  
      END IF;
  END PROCESS;
  
  s_rstrxrdy <= '1' WHEN (s_rxholdrd1='1' AND rxholdrd='0')    -- end of rx hold rd
                    ELSE '0';
  
  -- output data read from receiver holding register to host block
  rxdata <= s_rxholdreg;
  

  ------------------------------------------------------------------------------
  -- RECEIVER PARITY GENERATION
  ------------------------------------------------------------------------------
  -- initialises on start bit IF data is shifted in.
  -- EVEN parity, s_rxpty will be 0 for an even number of 1s in the received data.
  -- ODD  parity, s_rxpty will be 0 for an odd number of 1s in the received data.
  -- NO   parity, s_rxpty will always be 0.
  rx_ptygen : PROCESS (bclkx16,reseth,rxen,ptymode)
  BEGIN
    IF (reseth='1' OR rxen='0' OR ptymode=c_noparity) THEN
      s_rxpty  <='0';
    ELSIF (bclkx16'event AND bclkx16='1') THEN    
      IF (s_rxclken='1') THEN

        IF (s_pstate=start AND s_rxstart='1' AND s_rxshiftfull='0')
        THEN
          s_rxpty<=NOT(ptymode(0));
        ELSIF (s_rxaccept='1' AND s_pstate/=stop) THEN
          s_rxpty<=s_rxpty XOR rxd;                   -- calculate parity during shift in
        ELSE
          s_rxpty<=s_rxpty;
        END IF;
        
      END IF;      
      
    END IF;
    
  END PROCESS;

  ------------------------------------------------------------------------------
  -- FRAME ERROR CHECKING
  ------------------------------------------------------------------------------
  -- check stop bit is '1' at end of shift in. When the s_frmerr signal is '1', it
  -- indicates that the frame currently in the rx shift register has a frame error.
  rx_frmchk : PROCESS (reseth,rxen,bclkx16)
    BEGIN
      IF (reseth='1' OR rxen='0') THEN
        s_frmerr <= '0';
      ELSIF (bclkx16'event AND bclkx16='1') THEN
        IF (s_rxclken='1') THEN

          IF (s_pstate=stop AND s_rxaccept='1') THEN
            s_frmerr <= NOT(rxd);
          END IF;
          
        END IF;
      
      END IF;

  END PROCESS rx_frmchk; 
  
  
  
  ------------------------------------------------------------------------------
  -- RECEIVER ERROR FLAGS FOR STATUS REGISTER
  ------------------------------------------------------------------------------
  -- pty_err set to 1 when the calculated parity doesn't match the received parity
  -- bit (s_rxshiftreg(0)=0 and s_rxpty=1). It will remain set until a character
  -- without a parity error is received.
  -- frm_err set to 1 if the stop bit=0 (rxd=0 when s_rxshiftreg(0)=0).It will
  -- remain set until a character without a frame error is received.
  -- Both the parity & frame error flags are updated every time the holding register is
  -- loaded.
  -- ovr_err set if a new start bit has been accepted,but the rx shift register is full
  -- It will stay set until cleared  by the 'reset status' command.
  rx_errflags : PROCESS (bclkx16,reseth,rxen)
  BEGIN
    IF (reseth='1' OR rxen='0') THEN
      pty_err   <= '0';
      frm_err   <= '0';
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_rxclken='1') THEN

        -- set parity & frame error status reg flags on transfer of shift
        -- reg to holding reg
        IF ((s_pstate=stop AND s_rxaccept='1' AND s_rxrdy='0') OR
            (s_rxshiftfull='1' AND s_rxrdy='0'))                        
        THEN
          pty_err <= s_rxpty;
        END IF;
      
        IF (s_pstate=stop AND s_rxaccept='1' AND s_rxrdy='0') THEN
          frm_err <= NOT(rxd);
        ELSIF (s_rxshiftfull='1' AND s_rxrdy='0') THEN
          frm_err <= s_frmerr;
        END IF;
        
      END IF;
      
    END IF;
    
  END PROCESS;
  
  rx_ovrflag : PROCESS (bclkx16,reseth,rxen,rxstatusrst)
  BEGIN
    IF (reseth='1' OR rxen='0' OR rxstatusrst='1') THEN
      s_ovr_err <= '0';
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_rxclken='1') THEN
   
        -- overrun error, autosustained until cleared status reg clear cmd
        -- or receiver disabled or global reset
        IF ((s_rxshiftfull='1' AND s_pstate=start AND s_rxstart='1') OR
            (s_ovr_err='1')) 
        THEN
          s_ovr_err <= '1';
        END IF;
        
      END IF;

    END IF;
    
  END PROCESS;
  
  ovr_err <= s_ovr_err;
  
  
  

END ARCHITECTURE arc_receive;
-------------------------------------------------------------------------------
-- END OF FILE : RECEIVE.VHD
-------------------------------------------------------------------------------
