-------------------------------------------------------------------------------
-- FILE NAME   : TRANSMIT.VHD
-- PROJECT     : UART
-- VERSION     : A
-- AUTHOR      : M. S. HARVEY
-- SIM/SYN     : Synthesis
-- DESCRIPTION : Transmitter
-- DEPENDENCIES: STD_LOGIC_1164; NUMERIC_STD
-- LIMITATIONS :
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- CHANGE HISTORY.
-- REV.           DATE        DESCRIPTION					
--  A             05/12/99    Initial design.
-------------------------------------------------------------------------------



-- IEEE library
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;


ENTITY transmit IS
  PORT(
    hclk     : IN  std_logic;                      -- host clock
    reseth   : IN  std_logic;                      -- async reset
    bclkx16  : IN  std_logic;                      -- x16 baud rate clock
    txen     : IN  std_logic;                      -- transmitter enable
    cts      : IN  std_logic;                      -- clear to send
    txctsen  : IN  std_logic;                      -- transmitter inhibit by cts enable
    txholdwe : IN  std_logic;                      -- tx hold reg write enable
    data_in  : IN  std_logic_vector(7 DOWNTO 0);   -- data input for writes to tx hold reg
    ptymode  : IN  std_logic_vector(1 DOWNTO 0);   -- parity mode
    chanmode : IN  std_logic_vector(1 DOWNTO 0);   -- channel operating mode
    txempty  : OUT std_logic;                      -- transmitter hold & shift reg empty
    txrdy    : OUT std_logic;                      -- transmitter hold empty
    txd      : OUT std_logic                       -- transmitted serial data
        );
END ENTITY transmit;



ARCHITECTURE arc_transmit OF transmit IS


  -- internal signals
  SIGNAL s_txclken    : std_logic;                     -- transmitter clock enable
  SIGNAL s_txpty      : std_logic;                     -- transmitter parity
  SIGNAL s_txrdy      : std_logic;                     -- transmitter hold reg empty
  SIGNAL s_txempty    : std_logic;                     -- transmitter hold & shift reg empty
  SIGNAL s_txholdreg  : std_logic_vector(7 DOWNTO 0);  -- transmitter hold reg
  SIGNAL s_txshiftreg : std_logic_vector(9 DOWNTO 0);  -- transmitter shift reg
  SIGNAL s_txshiften  : std_logic;                     -- transmitter shift reg enable
  SIGNAL s_txshiften1 : std_logic;                     -- transmitter shift reg enable,delayed
  SIGNAL s_txshiften2 : std_logic;                     -- transmitter shift reg enable,delayed
  SIGNAL s_txholdwe1  : std_logic;                     -- transmitter hold reg write,delayed
  SIGNAL s_txcount    : unsigned(3 DOWNTO 0);

  -- parity modes
  CONSTANT c_noparity   : std_logic_vector(1 DOWNTO 0):= "00"; 
  CONSTANT c_evenparity : std_logic_vector(1 DOWNTO 0):= "01"; 
  CONSTANT c_oddparity  : std_logic_vector(1 DOWNTO 0):= "10";
  
  -- operating modes 
  CONSTANT c_normal    : std_logic_vector(1 DOWNTO 0):= "00"; 
  CONSTANT c_intloop   : std_logic_vector(1 DOWNTO 0):= "01"; 
  CONSTANT c_autoecho  : std_logic_vector(1 DOWNTO 0):= "10"; 
  

BEGIN

  ----------------------------------------------------------------------------
  -- TRANSMITTER CLOCK
  ----------------------------------------------------------------------------
  -- divide down the 16 times baud clock bclkx16 to provide the transmitter
  -- bit clock.
  tx_clkdiv : PROCESS (bclkx16, reseth,txen)
  
  BEGIN
    IF (reseth='1') THEN
      s_txcount<=(OTHERS =>'0');
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      s_txcount<= s_txcount + 1;   
    END IF;
  END PROCESS;

  s_txclken <= '1' WHEN (s_txcount=x"7") ELSE '0';

  ----------------------------------------------------------------------------
  -- TRANSMITTER HOLDING REGISTER
  ----------------------------------------------------------------------------
  -- can only be written if transmitter is enabled
  tx_hold : PROCESS (hclk,reseth)
  BEGIN
    IF (reseth='1') THEN
      s_txholdreg <= (OTHERS => '0');
    ELSIF (hclk'event AND hclk='1') THEN
      IF (txholdwe='1') THEN            -- write
        s_txholdreg<=data_in;
      END IF;          
    END IF;
  END PROCESS;

    
  ----------------------------------------------------------------------------
  -- TRANSMITTER PARITY GENERATION
  ----------------------------------------------------------------------------
  -- EVEN parity, s_txpty will be 0 for an even number of 1s in the transmitted data.
  -- ODD  parity, s_txpty will be 0 for an odd  number of 1s in the transmitted data.
  tx_ptygen : PROCESS (bclkx16, reseth)
  BEGIN
    IF (reseth='1') THEN
      s_txpty<='0';
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_txclken) = '1' THEN
      
        IF (s_txrdy='0' AND s_txshiften='0') THEN
          s_txpty<=NOT(ptymode(0));                -- initialise at start bit
        ELSE
          s_txpty<=s_txpty XOR s_txshiftreg(0);    -- calculate parity
        END IF;
        
      END IF;
      
    END IF;
  END PROCESS;
  
    
  ------------------------------------------------------------------------------
  -- TRANSMITTER SHIFT REGISTER
  ------------------------------------------------------------------------------
  -- Will load from the transmitter hold reg if finished shifting and data has
  -- been written to the transmitter hold reg and the transmitter is enabled.
  -- Starts shifting when s_txshiften goes active (high).Note that it will be
  -- shortened for 'no parity' mode.
  tx_shiftreg: PROCESS (bclkx16,reseth,txen)
  BEGIN
    IF (reseth='1') THEN 
      s_txshiften  <= '0';
      s_txshiftreg <= (OTHERS => '0');
      
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_txclken) = '1' THEN
     
        -- shift / load control.
        -- start shift after load from hold reg if not cts inhibited and transmitter enabled
        IF ((txctsen='0' OR (txctsen='1' AND cts='0')) 
          AND s_txrdy='0' AND s_txshiften='0' AND txen='1')
        THEN
          s_txshiften <= '1';                              -- start shift
        ELSIF (s_txshiftreg="0000000001") THEN             -- last bit being shifted
          s_txshiften <= '0';                              -- frame finished,disable shifting
        END IF;
        
        -- Transmitter shift register.
        -- Loads from hold reg if the hold reg has data (s_txrdy=0), the transmitter is
        -- enabled (txen=1), the previous shift has finished (s_txshiften=0).
        -- Starts shifting on txclk edge after load conditions are valid, will continue
        -- shifting even if transmitter is disabled during shift.
        IF (s_txrdy='0' AND s_txshiften='0' AND txen='1') THEN
          s_txshiftreg(9 DOWNTO 0) <= "11" & s_txholdreg;  -- load data byte to be transmitted
        ELSIF (s_txshiften='1') THEN
          s_txshiftreg(9) <= '0';                          -- shift 0 into ms bit
          s_txshiftreg(8) <= s_txshiftreg(9);
          IF (ptymode = c_noparity) THEN
            s_txshiftreg(7) <= s_txshiftreg(9);            -- shorten for no parity mode
          ELSE
            s_txshiftreg(7) <= s_txshiftreg(8);
          END IF;
          s_txshiftreg(6 DOWNTO 0) <= s_txshiftreg(7 DOWNTO 1);
        END IF;
        
      END IF;   
      
    END IF;
  END PROCESS;
    
  
  
  ------------------------------------------------------------------------------
  -- TRANSMITTER STATUS BITS
  ------------------------------------------------------------------------------
  -- detect rising edge of write to tx holding reg, set txrdy & txempty
  tx_status : PROCESS (hclk,reseth)
  BEGIN
    IF (reseth='1') THEN
      s_txholdwe1  <= '0';
      s_txshiften1 <= '0';
      s_txshiften2 <= '0';
      s_txrdy      <= '1';
      s_txempty    <= '1';
    ELSIF (hclk'event AND hclk = '1') THEN
      s_txholdwe1  <= txholdwe;
      s_txshiften1 <= s_txshiften;
      s_txshiften2 <= s_txshiften1;
      
      -- tx_rdy control
      IF (s_txholdwe1='1') THEN
        s_txrdy <= '0';                                        -- clear after write to buffer
      ELSIF (s_txshiften1='1' AND s_txshiften2='0') THEN
        s_txrdy <= '1';                                        -- set on start of shift
      END IF;
      
      -- tx_empty control
      IF (s_txholdwe1='1') THEN
        s_txempty <= '0';                                       -- clear after write to buffer
      ELSIF (s_txrdy='1' AND s_txshiften1='0' AND s_txshiften2='1') THEN
        s_txempty <= '1';                                       -- set at end of shift if hold reg empty
      END IF;
      
    END IF;
  END PROCESS;
  
  -- txrdy & txempty are forced to 0 for autoecho mode
  txrdy   <= s_txrdy   WHEN (chanmode/=c_autoecho) ELSE '0';
  txempty <= s_txempty WHEN (chanmode/=c_autoecho) ELSE '0';
  
  ------------------------------------------------------------------------------
  -- TRANSMIT SERIAL DATA
  ------------------------------------------------------------------------------
  tx_data: PROCESS (bclkx16,reseth)
  BEGIN
    IF (reseth='1') THEN 
      txd <= '1';
    ELSIF (bclkx16'event AND bclkx16='1') THEN
      IF (s_txclken) = '1' THEN
    
        CASE chanmode IS
          
          -- normal & internal loopback modes
          WHEN c_normal|c_intloop =>
          
            IF ((txctsen='0' OR (txctsen='1' AND cts='0')) 
                      AND s_txrdy='0' AND s_txshiften='0' AND txen='1')
            THEN
              txd <= '0';                                           -- transmit start bit
            ELSIF (s_txshiften='1' AND s_txshiftreg="0000000011" 
                   AND (ptymode=c_evenparity OR ptymode=c_oddparity)) 
            THEN
              txd <= s_txpty;                                       -- transmit parity bit
            ELSIF (s_txshiften='1') THEN
              txd <= s_txshiftreg(0);                               -- output from shift reg
            ELSE
              txd <= '1';                                           -- transmit mark
            END IF;
          
          -- autoecho mode  
          WHEN OTHERS =>
            txd <= '1';
            
        END CASE;
        
      END IF;
              
    END IF;
  END PROCESS;
      

END ARCHITECTURE arc_transmit;
-------------------------------------------------------------------------------
-- END OF FILE : TRANSMIT.VHD
-------------------------------------------------------------------------------
