-------------------------------------------------------------------------------
-- FILE NAME   : UART.VHD
-- PROJECT     : UART
-- VERSION     : A
-- AUTHOR      : M. S. HARVEY
-- SIM/SYN     : Synthesis
-- DESCRIPTION : UART top level
-- DEPENDENCIES: IEEE.STD_LOGIC_1164
-- LIMITATIONS :
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- CHANGE HISTORY.
-- REV.           DATE        DESCRIPTION					
--  A             02/12/99    Initial design.
-------------------------------------------------------------------------------



-- IEEE library
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;


ENTITY uart IS
  PORT(
    hclk   : IN     std_logic;                         -- host clock
    reseth : IN     std_logic;                         -- async reset
    cs     : IN     std_logic;                         -- chip select
    rw     : IN     std_logic;                         -- read/write
    addr   : IN     std_logic_vector(2 DOWNTO 0);      -- host address
    data   : INOUT  std_logic_vector(7 DOWNTO 0);      -- host data
    ack    : OUT    std_logic;                         -- transfer acknowledge to host
    irq    : OUT    std_logic;                         -- interrupt request to host
    cts    : IN     std_logic;                         -- clear to send
    rts    : OUT    std_logic;                         -- ready to send
    rxd    : IN     std_logic;                         -- received serial data
    txd    : OUT    std_logic                          -- transmitted serial data
        );
END ENTITY uart;



ARCHITECTURE arc_uart OF uart IS

  -- component declarations for lower level design entities
  
  COMPONENT host
    PORT(
      hclk       : IN  std_logic;                      -- host clock
      reseth     : IN  std_logic;                      -- async reset
      cs         : IN  std_logic;                      -- chip select
      rw         : IN  std_logic;                      -- read/write
      addr       : IN  std_logic_vector(2 DOWNTO 0);   -- host address
      data_in    : IN  std_logic_vector(7 DOWNTO 0);   -- host data input for writes
      rxdata     : IN  std_logic_vector(7 DOWNTO 0);   -- received data
      frm_err    : IN  std_logic;                      -- frame error from receiver
      pty_err    : IN  std_logic;                      -- parity error from receiver
      ovr_err    : IN  std_logic;                      -- overrun error from receiver
      txempty    : IN  std_logic;                      -- transmitter hold & shift reg empty
      txrdy      : IN  std_logic;                      -- transmitter hold reg empty
      rxrdy      : IN  std_logic;                      -- receiver data ready
      rxfull     : IN  std_logic;                      -- receiver hold & shift reg full
      n          : OUT std_logic_vector(15 DOWNTO 0);  -- brsel0,1 concatenated
      ptymode    : OUT std_logic_vector(1 DOWNTO 0);   -- parity mode
      chanmode   : OUT std_logic_vector(1 DOWNTO 0);   -- channel operating mode
      txen       : OUT std_logic;                      -- transmitter enable
      txctsen    : OUT std_logic;                      -- transmitter inhibit by cts enable
      rxrtsen    : OUT std_logic;                      -- receiver rts enable
      rxen       : OUT std_logic;                      -- receiver enable
      txholdwe   : OUT std_logic;                      -- transmitter hold reg write enable
      rxholdrd   : OUT std_logic;                      -- receiver hold reg read enable
      rxstatusrst: OUT std_logic;                      -- status register reset (rx flags)
      data_out   : OUT std_logic_vector(7 DOWNTO 0);   -- host data output for reads
      dataoe     : OUT std_logic;                      -- host data output enable for reads
      irq        : OUT std_logic;                      -- interrupt request to host
      ack        : OUT std_logic                       -- transfer acknowledge to host
          );
  END COMPONENT;
  
  COMPONENT baudgen
    GENERIC (
      accum_size : integer:=16);                       -- accumulator size (in bits)
    PORT(
      hclk     : IN  std_logic;                        -- host clock
      reseth   : IN  std_logic;                        -- async reset
      n        : IN  std_logic_vector(15 DOWNTO 0);    -- BRSEL1,0
      bclkx16  : OUT std_logic                         -- baud clock x 16
      );
  END COMPONENT;
  
  
  COMPONENT transmit
  PORT(
    hclk     : IN  std_logic;                          -- host clock
    reseth   : IN  std_logic;                          -- async reset
    bclkx16  : IN  std_logic;                          -- x16 baud rate clock
    txen     : IN  std_logic;                          -- transmitter enable
    cts      : IN  std_logic;                          -- clear to send
    txctsen  : IN  std_logic;                          -- transmitter inhibit by cts enable
    txholdwe : IN  std_logic;                          -- tx hold reg write enable
    data_in  : IN  std_logic_vector(7 DOWNTO 0);       -- data input for writes to tx hold reg
    ptymode  : IN  std_logic_vector(1 DOWNTO 0);       -- parity mode
    chanmode : IN  std_logic_vector(1 DOWNTO 0);       -- channel operating mode
    txempty  : OUT std_logic;                          -- transmitter hold & shift reg empty
    txrdy    : OUT std_logic;                          -- transmitter hold empty
    txd      : OUT std_logic                           -- transmitted serial data
        );
  END COMPONENT;

    
  COMPONENT receive
    PORT(
      hclk       : IN  std_logic;                      -- host clock
      reseth     : IN  std_logic;                      -- async reset
      bclkx16    : IN  std_logic;                      -- x16 baud rate clock
      rxen       : IN  std_logic;                      -- receiver enable
      rxd        : IN  std_logic;                      -- received data
      rxrtsen    : IN  std_logic;                      -- rts enable
      rxholdrd   : IN  std_logic;                      -- rx hold reg read enable
      rxstatusrst: IN  std_logic;                      -- status register clear command
      ptymode    : IN  std_logic_vector(1 DOWNTO 0);   -- parity mode
      rxdata     : OUT std_logic_vector(7 DOWNTO 0);   -- data output for reads from rx hold reg
      rts        : OUT std_logic;                      -- ready to send
      rxrdy      : OUT std_logic;                      -- receiver hold reg full
      rxfull     : OUT std_logic;                      -- receiver hold & shift reg full
      frm_err    : OUT std_logic;                      -- frame error to host
      pty_err    : OUT std_logic;                      -- parity error to host
      ovr_err    : OUT std_logic                       -- overrun error to host
        );
  END COMPONENT;

  -- internal signals
  SIGNAL s_frm_err     : std_logic;
  SIGNAL s_pty_err     : std_logic;
  SIGNAL s_ovr_err     : std_logic;
  SIGNAL s_txempty     : std_logic;
  SIGNAL s_txrdy       : std_logic;
  SIGNAL s_rxrdy       : std_logic;
  SIGNAL s_rxfull      : std_logic;
  SIGNAL s_n           : std_logic_vector(15 DOWNTO 0);
  SIGNAL s_ptymode     : std_logic_vector(1 DOWNTO 0);
  SIGNAL s_chanmode    : std_logic_vector(1 DOWNTO 0);
  SIGNAL s_txen        : std_logic;
  SIGNAL s_txctsen     : std_logic;
  SIGNAL s_rxrtsen     : std_logic;
  SIGNAL s_rxen        : std_logic;
  SIGNAL s_txholdwe    : std_logic;
  SIGNAL s_rxholdrd    : std_logic;
  SIGNAL s_rxstatusrst : std_logic;
  SIGNAL s_data_in     : std_logic_vector(7 DOWNTO 0);
  SIGNAL s_data_out    : std_logic_vector(7 DOWNTO 0);
  SIGNAL s_rxdata      : std_logic_vector(7 DOWNTO 0);
  SIGNAL s_dataoe      : std_logic;
  SIGNAL s_rts         : std_logic;
  SIGNAL s_cts         : std_logic;
  SIGNAL s_txd         : std_logic;
  SIGNAL s_rxd         : std_logic;
  SIGNAL s_bclkx16     : std_logic;

  -- parity modes
  CONSTANT c_noparity   : std_logic_vector(1 DOWNTO 0):= "00"; 
  CONSTANT c_evenparity : std_logic_vector(1 DOWNTO 0):= "01"; 
  CONSTANT c_oddparity  : std_logic_vector(1 DOWNTO 0):= "10";
  
  -- operating modes 
  CONSTANT c_normal    : std_logic_vector(1 DOWNTO 0):= "00"; 
  CONSTANT c_intloop   : std_logic_vector(1 DOWNTO 0):= "01"; 
  CONSTANT c_autoecho  : std_logic_vector(1 DOWNTO 0):= "10"; 


BEGIN
  
  ----------------------------------------------------------------------------
  -- HOST INTERFACE
  ----------------------------------------------------------------------------
  U1 : host
    PORT MAP(
      hclk       => hclk,
      reseth     => reseth,
      cs         => cs,
      rw         => rw,
      addr       => addr,
      data_in    => s_data_in,
      rxdata     => s_rxdata,
      frm_err    => s_frm_err,
      pty_err    => s_pty_err,
      ovr_err    => s_ovr_err,
      txempty    => s_txempty,
      txrdy      => s_txrdy,
      rxrdy      => s_rxrdy,
      rxfull     => s_rxfull,
      n          => s_n,
      ptymode    => s_ptymode,
      chanmode   => s_chanmode,
      txen       => s_txen,
      txctsen    => s_txctsen,
      rxrtsen    => s_rxrtsen,
      rxen       => s_rxen,
      txholdwe   => s_txholdwe,
      rxholdrd   => s_rxholdrd,
      rxstatusrst=> s_rxstatusrst,
      data_out   => s_data_out,
      dataoe     => s_dataoe,
      irq        => irq,
      ack        => ack);


  ----------------------------------------------------------------------------
  -- BAUD RATE GENERATION
  ----------------------------------------------------------------------------  
  U2 : baudgen
    GENERIC MAP(
      accum_size => 16)           -- must be equal to or greater than width of n
    PORT MAP(
      hclk     => hclk,
      reseth   => reseth,
      n        => s_n,
      bclkx16  => s_bclkx16);
                    
                    
  ----------------------------------------------------------------------------
  -- TRANSMITTER
  ----------------------------------------------------------------------------
  U3 : transmit
    PORT MAP(
      hclk     => hclk,
      reseth   => reseth,
      bclkx16  => s_bclkx16,
      txen     => s_txen,
      cts      => s_cts,
      txctsen  => s_txctsen,
      txholdwe => s_txholdwe,
      data_in  => s_data_in,
      ptymode  => s_ptymode,
      chanmode => s_chanmode,
      txempty  => s_txempty,
      txrdy    => s_txrdy,
      txd      => s_txd);
      
      -- transmitted serial data
      WITH s_chanmode SELECT
        txd <= s_txd WHEN c_normal,      -- transmit data from tx shift reg
               rxd   WHEN c_autoecho,    -- retransmit received data
               '1'   WHEN OTHERS;        -- internal loopback mode
      
      -- s_cts is either the cts input from the external receiver or the
      -- rts signal from the internal receiver during internal loopback mode
      s_cts <= s_rts WHEN (s_chanmode = c_intloop) ELSE cts;

                           
  ----------------------------------------------------------------------------
  -- RECEIVER
  ----------------------------------------------------------------------------
  U4 : receive
    PORT MAP(
      hclk       => hclk,
      reseth     => reseth,
      bclkx16    => s_bclkx16,
      rxen       => s_rxen,
      rxd        => s_rxd,
      rxrtsen    => s_rxrtsen,
      rxholdrd   => s_rxholdrd,
      rxstatusrst=> s_rxstatusrst,
      ptymode    => s_ptymode,
      rxdata     => s_rxdata,
      rxfull     => s_rxfull,
      rts        => s_rts,
      rxrdy      => s_rxrdy,
      frm_err    => s_frm_err,
      pty_err    => s_pty_err,
      ovr_err    => s_ovr_err);
  
      -- ready to send output - not active for internal loopback
      rts <= '1'  WHEN (s_chanmode=c_intloop) ELSE s_rts;

      -- s_rxd is either the rxd input or the txd input from the
      -- transmitter for internal loopback mode
      s_rxd <= s_txd WHEN (s_chanmode = c_intloop) ELSE rxd;

  ----------------------------------------------------------------------------
  -- DATA BUS BUFFERING
  ----------------------------------------------------------------------------
  data <= s_data_out WHEN (s_dataoe='0') ELSE (OTHERS => 'Z');
  s_data_in <= data;
  
  
  

END ARCHITECTURE arc_uart;
-------------------------------------------------------------------------------
-- END OF FILE : UART.VHD
-------------------------------------------------------------------------------
